# GraphQL Directives

A directive-based approach to GraphQL schema implementations. Provides a
"Directable" schema plugin that loads a GraphQL schema definition file and
implements by executing directive definitions.

## Usage

### Hello world!

Create a GraphQL schema definition file, and annotate it with directives.

```graphql
type Query {
  hello: String @value(string: "Hello world!")
}
```

Configure a GraphQL server with the "Directable" schema plugin, and set the
schema definition path to the created schema file. Now you should be able to run
the query `{ hello }` in the GraphQL explorer.

Most editors and IDE's support the
[GraphQL config](https://the-guild.dev/graphql/config) format. Create a file
like this in the Drupal folder and you should immedately get autocompletion and
type checking for available directives:

```json
{
  "$schema": "https://unpkg.com/graphql-config@latest/config-schema.json",
  "schema": ["./graphql/schema.gql", "./modules/*/directives.gql"],
  "documents": ["./graphql/queries/**/*.gql"]
}
```

This module itself makes use of that file as well. Any `directives.gql` files in
enabled modules are attached to the schema automatically, but it will also look
for any files that match the `schema` key in this GraphQL config file.

### Chaining

Directives can be chained to combine reusable data producers. They are composed
from left to right, meaning the output of the left directive is passed as parent
value to its right neighbour.

```graphql
type Query {
  # This will emit "three".
  list: String! @value(json: "[\"one\", \"two\", \"three\"]") @seek(pos: 2)
}
```

### Mapping

The `@map` directive allows to map over the output of its left neighbour and
apply all directives on the right side to each item.

```graphql
type Query {
  # This will emit ["a", "b"].
  map: [String!]!
    @value(json: "[{\"x\": \"a\"},{\"x\": \"b\"}]")
    @map
    @prop(key: "x")
}
```

### Default values

Since Drupal's data structures can't guarantee integrity, the graphql schema
will enforce default values as much as possible. Whenever a type is used in a
non-nullable position (no `!` at the end ), it attempts to apply a default value
if the value is `null`. The default value is determined by the type, e.g. `0`
for `Int`, `false` for `Boolean`, `""` for `String` and `[]` for list types like
`[String!]!`.

For custom types, interface, unions or scalars, the `@default` directive can be
used to start a directive chain that generates a default value.

```graphql
scalar MyScalar @default @value(string: "bar")

type Query {
  # This will emit `''`.
  string: String! @value
  # This will emit `0`.
  int: Int! @value
  # This will emit `[]`.
  list: [String!]! @value
  # This will emit `bar`
  manual: MyScalar! @value
}
```

### Type resolution

Directives can also be used to resolve the runtime types of unions and
interfaces. To do that, apply any directives that can be used to resolve field
values to the interface or union. The chain of directives should resolve to a
string value which will be treated as a type id.

```graphql
union Letters @prop(key: "type") = A | B
```

This resolved type id will then be matched against object types annotated with
the `@type` directive to retrieve the actual type.

```graphql
type A @type(id: "a") {
  type: String!
}

type B @type(id: "b") {
  type: String!
}
```

## Argument handling

Directive arguments should always be of type `String`. If the directive argument
equals `$`, the current value will be passed as the argument value. If its `$`,
followed by any characters, these characters will be used as a key to retrieve
the value from the current field arguments.

```graphql
type Query {
  static: Post @loadEntity(type: "node", id: "1")
  parent: Post @value(int: 1) @loadEntity(type: "node", id: "$")
  argument(id: String!): Post @loadEntity(type: "node", id: "$id")
}
```

## Built-in directives

The module ships with a variety of directives to accomplish common tasks. For a
comprehensive list consult `directives.gql` or editors LSP integration 😎.

## Extending

To add custom directives, create a module and place a `directives.gql` file at
it's root folder. This file should contain any new directives, annotated with
implementation hints:

```graphql
"""
Echo any message that is passed in.

implementation(drupal): Drupal\echo\Directives::echo
"""
directive @echo(message: String!) on FIELD_DEFINITION
```

The line starting with `implementation(...): ...` will be used by to determine
which PHP function to execute for this directive. It can be either a static
class method (`\Drupal\echo\Directives::echo`) or a Symfony service method
(`echo_service::echo`). Either way this function has to conform to one of two
possible interfaces:

### Simple directives

The preferred way of implementing new directives. The function has to accept a
single argument of type `\Drupal\graphql_directives\Api` and return any defined
type. The `Api` object provides access to the `parent` value, pre-resolved
`args` and the GraphQL field context, which can be used to interact with the
context and manage cache meta-information:

```php
<?php

use \Drupal\graphql_directives\Api;

class Directives {
  public static function echo(Api $api): string {
    return $api->args['message'];
  }
}
```

### Backwards compatible directives

There are already a lot of existing data producers for the GraphQL module for
Drupal. If a directive needs to re-use any of them, it can do so by adhering to
a different function interface. If it accepts a single argument of type
`\Drupal\graphql\GraphQL\ResolverBuilder` and returns an object of type
`\Drupal\graphql\GraphQL\Resolver\ResolverInterface`. In this case it operates
in "legacy" mode and allows to use the classic `ResolverBuilder` approach to
assemble data producers:

```php
<?php

use \Drupal\graphql\GraphQL\ResolverBuilder;
use \Drupal\graphql\GraphQL\Resolver\ResolverInterface;

class Directives {
  public static function echo(ResolverBuilder $builder): ResolverInterface {
    return $builder->fromArgument('message');
  }
}
```

There is one notable difference! `$builder->fromArgument(...)` will return the
aliased value of the directive argument (see "Argument handling" above) instead
of the actual field argument.s

### Execution contexts

One detail we did not mention so far is the `drupal` in the
`implementation(drupal): ...` annotation. That's the "execution context". Each
directive can have multiple implementations, with different execution contexts.
This allows to swap out implementations of functionality based on the
circumstances. The schema configuration form has a textfield to define the
instances context, which defaults to Drupal.

One use case that this solved was swapping out image handling based on the
environment:

```graphql
"""
Generate the public url for a given processed image.

Using gd and Drupal image presets:
implementation(drupal): \Drupal\custom_images\DrupalImage::url

Using cloudinary:
implementation(drupal,cloudinary): \Drupal\custom_images\CloudinaryImage::url

Using a local thumbor instance:
implementation(drupal,thumbor): \Drupal\custom_images\ThumborImage::url
"""
directive @imageUrl(preset: String!) on FIELD_DEFINITION
```

## Schema Extensions

The module provides a `DirectableSchemaExtensionPluginBase` class that can be
used to create schema extensions that react to directives in the parent schema
definition. A schema extension plugin for the Drupal GraphQL module provides two
schema definitions: one for the "base" schema and one for the actual extensions.
In case of directable schema extension, the base schema definition should
contain the directives while the extension schema defines the derived types and
fields.

For a very simple example, please refer to the `graphql_directives_test` module.
