<?php

namespace Drupal\graphql_directives\Directives;

use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\TypedData\TranslatableInterface;
use Drupal\graphql\GraphQL\ResolverBuilder;
use Drupal\graphql\GraphQL\Resolver\ResolverInterface;

/**
 * Resolvers related to entities.
 */
class Entities {

  /**
   * Pull a specific typed-data property from an entity.
   */
  public static function property(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('property_path', [
      'path' => $builder->fromArgument('path'),
      'value' => $builder->fromParent(),
      'type' => $builder->callback(
        fn($entity) => $entity->getTypedData()->getDataDefinition()->getDataType()
      ),
    ]);
  }

  /**
   * Resolve an entity reference field.
   */
  public static function entityReference(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_reference')
      ->map('entity', $builder->fromParent())
      ->map('language', $builder->compose(
        $builder->fromParent(),
        $builder->callback(fn (TranslatableInterface $value) => $value->language()->getId())
      ))
      ->map('field', $builder->fromArgument('field'));
  }

  /**
   * Resolve an entity reference revisions field.
   */
  public static function entityReferenceRevisions(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_reference_revisions')
      ->map('entity', $builder->fromParent())
      ->map('language', $builder->compose(
        $builder->fromParent(),
        $builder->callback(fn(TranslatableInterface $value) => $value->language()->getId())
      ))
      ->map('field', $builder->fromArgument('field'));
  }

  /**
   * Retrieve a specific translation of an entity.
   */
  public static function entityTranslation(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_translation')
      ->map('entity', $builder->fromParent())
      ->map('language', $builder->fromArgument('lang'));
  }

  /**
   * Retrieve all translations of an entity.
   */
  public static function entityTranslations(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_translations')
      ->map('entity', $builder->fromParent());
  }

  /**
   * Retrieve an entities bundle.
   */
  public static function entityBundle(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_bundle')
      ->map('entity', $builder->fromParent());
  }

  /**
   * Retrieve an entities id.
   */
  public static function entityId(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_id')
      ->map('entity', $builder->fromParent());
  }

  /**
   * Retrieve an entities label.
   */
  public static function entityLabel(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_label')
      ->map('entity', $builder->fromParent());
  }

  /**
   * Retrieve an entities type id.
   */
  public static function entityType(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_type_id')->map('entity', $builder->fromParent());
  }

  /**
   * Retrieve an entities url path.
   */
  public static function entityPath(ResolverBuilder $builder) : ResolverInterface {
    return $builder->compose(
      $builder->produce('entity_url')->map('entity', $builder->fromParent()),
      $builder->produce('url_path')->map('url', $builder->fromParent()),
      $builder->produce('sanitize_path')->map('raw', $builder->fromParent())
    );
  }

  /**
   * Retrieve an entities uuid.
   */
  public static function entityUuid(ResolverBuilder $builder) : ResolverInterface {
    return $builder->produce('entity_uuid')
      ->map('entity', $builder->fromParent());
  }

  /**
   * Retrieve the language of an entity.
   *
   * Use the language path if it is configured or fallback to the language ID.
   * Format the output so it's compliant with GraphQL enums.
   */
  public static function entityLanguage(ResolverBuilder $builder) : ResolverInterface {
    return $builder->compose(
      $builder->produce('entity_language')->map('entity', $builder->fromParent()),
      $builder->callback(function (LanguageInterface $lang) {
        $prefixes = \Drupal::config('language.negotiation')->get('url.prefixes');
        $languagePath = !empty($prefixes[$lang->getId()]) ? $prefixes[$lang->getId()] : $lang->getId();
        // Make it compliant with GraphQL enums,
        // so convert '-' delimiter to '_'.
        // Examples:
        // - Language ID: gsw-berne, zh-hans
        // - Language path (ISO 639-1): de-CH, zh-CN
        $languagePath = str_replace('-', '_', $languagePath);
        return $languagePath;
      })
    );
  }

  /**
   * Load a given entity by it's path or type and id or uuid.
   */
  public static function loadEntity(ResolverBuilder $builder) : ResolverInterface {
    $operation = $builder->fromArgument('operation') ?? $builder->fromValue('view');

    if ($type = $builder->fromArgument('type')) {
      if ($uuid = $builder->fromArgument('uuid')) {
        return $builder->produce('entity_load_by_uuid')
          ->map('type', $type)
          ->map('uuid', $uuid)
          ->map('access_operation', $operation);
      }
      elseif ($id = $builder->fromArgument('id')) {
        return $builder->produce('entity_load')
          ->map('type', $type)
          ->map('id', $id)
          ->map('access_operation', $operation);
      }
    }

    if ($route = $builder->fromArgument('route') ?? $builder->fromParent()) {
      return $builder->produce('route_entity')
        ->map('url', $route)
        ->map('language', $builder->produce('url_language')->map('url', $route));
    }
    throw new \Exception('Invalid use of @loadEntity. Provide either a route or type and id/uuid.');
  }

}
