<?php

namespace Drupal\graphql_directives;

use Drupal\Component\DependencyInjection\ContainerInterface;
use Drupal\graphql\GraphQL\ResolverBuilder;
use Drupal\graphql\GraphQL\Resolver\ResolverInterface;
use GraphQL\Utils\BuildSchema;

/**
 * Extracts directive implementation details from a GraphQL schema.
 */
class ImplementationExtractor {

  public function __construct(protected ContainerInterface $container) {}

  /**
   * Parses the schema and extracts the directive implementations.
   *
   * @param \GraphQL\Language\AST\DocumentNode | string $schema
   *   The parsed schema definition document.
   * @param array<int,string> $contexts
   *   Contexts that are relevant for this implementation.
   *
   * @return array<string, array<string, string>|null>
   *   An array mapping directive names to their implementation details
   *   (service/class and method) or NULL if no suitable implementation
   *   is found.
   */
  public function extract(mixed $schema, array $contexts = []): array {
    $parsed = BuildSchema::build($schema);
    $directives = [];
    $skip = [
      'skip',
      'include',
      'deprecated',
    ];
    foreach ($parsed->getDirectives() as $directive) {
      // Skip directives that are not relevant.
      if (in_array($directive->name, $skip)) {
        continue;
      }

      if (preg_match_all('/implementation\(([^)]+)\): ([^:]+)::([^\s]+)/', $directive->description ?? '', $matches, PREG_SET_ORDER)) {
        $implementations = array_map(function ($match) {
          return [
            "contexts" => array_map('trim', explode(",", $match[1])),
            strpos($match[2], '\\') === 0 ? 'class' : 'service' => $match[2],
            'method' => $match[3],
          ];
        }, $matches);

        // Find the implementation with the biggest overlap of contexts.
        $bestImplementation = NULL;
        $maxOverlap = 0;
        foreach ($implementations as $implementation) {
          $overlap = count(array_intersect($implementation['contexts'], $contexts));
          // Skip if there's no overlap at all.
          if ($overlap === 0) {
            continue;
          }
          // Select if it has more overlap, or if it's the first one with the same overlap.
          if ($overlap > $maxOverlap || ($overlap === $maxOverlap && $bestImplementation === NULL)) {
            unset($implementation['contexts']);
            $bestImplementation = $implementation;
            $maxOverlap = $overlap;
          }
        }

        $directives[$directive->name] = $bestImplementation;
      }
      else {
        // Store NULL if no implementation details found in the description.
        $directives[$directive->name] = NULL;
      }
      foreach ($directives as $name => $directive) {
        if (!$directive) {
          continue;
        }
        if (isset($directive['class'])) {
          $class = $directive['class'];
        }
        if (isset($directive['service'])) {
          $class = get_class($this->container->get($directive['service']));
        }
        if ($class === DeprecatedPluginCompatibility::class) {
          $directives[$name]['legacy'] = TRUE;
          continue;
        }
        $reflection = new \ReflectionMethod($class, $directive['method']);
        $exception = new MethodInterfaceException($class, $directive['method']);
        $parameters = $reflection->getParameters();
        $inputType = count($parameters) > 0 ? $parameters[0]->getType()->__toString() : NULL;
        $returnType = $reflection->hasReturnType() ? $reflection->getReturnType()->__toString() : NULL;
        if (count($parameters) > 1) {
          throw $exception;
        }
        if (!$inputType || $inputType === Api::class) {
          if (!$reflection->hasReturnType()) {
            throw $exception;
          }
        }
        elseif ($inputType === ResolverBuilder::class) {
          if ($returnType !== ResolverInterface::class) {
            throw $exception;
          }
          $directives[$name]['legacy'] = TRUE;
        }
        else {
          throw $exception;
        }
      }
    }
    // Filter out directives that ended up with no suitable implementation.
    return array_filter($directives, fn($directive) => $directive !== NULL);
  }

}
