<?php

namespace Drupal\graphql_directives\Plugin\GraphQL\Schema;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\graphql\GraphQL\Execution\FieldContext;
use Drupal\graphql\GraphQL\ResolverBuilder;
use Drupal\graphql\GraphQL\ResolverRegistry;
use Drupal\graphql\Plugin\GraphQL\Schema\ComposableSchema;
use Drupal\graphql\Plugin\SchemaExtensionPluginManager;
use Drupal\graphql_directives\DirectableSchemaExtensionPluginBase;
use Drupal\graphql_directives\DirectiveInterpreter;
use Drupal\graphql_directives\ImplementationExtractor;
use Drupal\graphql_directives\SchemaAggregator;
use GraphQL\Language\AST\DocumentNode;
use GraphQL\Language\AST\ObjectTypeDefinitionNode;
use Symfony\Component\DependencyInjection\ContainerInterface as SymfonyContainerInterface;

/**
 * A directive base schema plugin.
 *
 * @Schema(
 *   id = "directable",
 *   name = "Directable schema"
 * )
 */
class DirectableSchema extends ComposableSchema {

  protected $typeMap = [];

  protected SchemaAggregator $aggregator;

  /**
   * {@inheritdoc}
   *
   * @codeCoverageIgnore
   */
  public static function create(SymfonyContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('cache.graphql.ast'),
      $container->get('module_handler'),
      $container->get('plugin.manager.graphql.schema_extension'),
      $container->get('graphql_directives.aggregator'),
      $container->getParameter('graphql.config')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $pluginId,
    array $pluginDefinition,
    ?CacheBackendInterface $astCache,
    ?ModuleHandlerInterface $moduleHandler,
    ?SchemaExtensionPluginManager $extensionManager,
    SchemaAggregator $aggregator,
    array $config,
  ) {
    $this->aggregator = $aggregator;

    parent::__construct(
      $configuration,
      $pluginId,
      $pluginDefinition,
      $astCache,
      $moduleHandler,
      $extensionManager,
      $config
    );
  }

  /**
   * Retrieves the raw schema definition string.
   *
   * @return string
   *   The schema definition.
   *
   * @throws \Exception
   */
  public function getSchemaDefinition(): string {
    $cid = implode(self::DERIVATIVE_SEPARATOR, ['schema_definition', $this->getPluginId()]);
    $cache = $this->astCache->get($cid);
    if (!$this->inDevelopment && $cache) {
      return $cache->data;
    }
    $file = $this->configuration['schema_definition'];
    if (!file_exists($file)) {
      throw new \Exception("Schema file '$file' could not be found.");
    }
    $schemaFiles = $this->aggregator->aggregate($file);
    $schemaParts = [];
    foreach ($schemaFiles as $file) {
      $schemaParts[] = file_get_contents($file);
    }
    $schemaString = implode("\n", $schemaParts);
    $this->astCache->set($cid, $schemaString);
    return $schemaString;
  }

  /**
   * Retrieve all autoload contexts and default to "drupal".
   */
  protected function autoloadContexts() : array {
    $contexts = $this->configuration['autoload_contexts'];
    if (!$contexts) {
      $contexts = 'drupal';
    }
    return array_map('trim', explode(',', $contexts));
  }

  /**
   * Create a per-instance plugin id.
   */
  public function getPluginId() : string {
    return implode(self::DERIVATIVE_SEPARATOR, [
      parent::getPluginId(),
      hash('sha256', $this->configuration['schema_definition'] . ':' . $this->configuration['autoload_contexts']),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getResolverRegistry() : ResolverRegistry {
    $registry = new ResolverRegistry();
    $extensions = $this->getExtensions();
    $builder = new ResolverBuilder();
    $document = $this->getSchemaDocument($extensions);

    foreach ($document->definitions as $definition) {
      if ($definition instanceof ObjectTypeDefinitionNode) {
        foreach ($definition->directives as $directive) {
          if ($directive->name->value === 'type' && $directive->arguments) {
            foreach ($directive->arguments as $argument) {
              if ($argument->name->value === 'id') {
                $this->typeMap[$argument->value->value] = $definition->name->value;
              }
            }
          }
        }
      }
    }

    $contexts = $this->autoloadContexts();

    $fieldResolvers = [];
    $typeResolvers = [];

    $autoloadCacheId = implode(self::DERIVATIVE_SEPARATOR, [
      $this->getPluginId(),
      'autoload',
      hash('sha256', $this->getSchemaDefinition()),
    ]);
    $autoloadCache = $this->astCache?->get($autoloadCacheId);
    if ($autoloadCache) {
      $autoload = $autoloadCache->data ?? [];
    }
    else {
      $autoload = (new ImplementationExtractor(\Drupal::getContainer()))->extract($document, $contexts);
      $this->astCache?->set($autoloadCacheId, $autoload, CacheBackendInterface::CACHE_PERMANENT, ['graphql_schema']);
    }

    $resolversCacheId = implode(self::DERIVATIVE_SEPARATOR, [
      $this->getPluginId(),
      'resolvers',
      hash('sha256', $this->getSchemaDefinition()),
    ]);
    $resolversCache = $this->astCache?->get($resolversCacheId);
    if ($resolversCache) {
      $fieldResolvers = $resolversCache->data['field_resolvers'] ?? [];
      $typeResolvers = $resolversCache->data['type_resolvers'] ?? [];
    }
    else {
      $interpreter = new DirectiveInterpreter($document, $builder, \Drupal::getContainer(), $autoload);
      $interpreter->interpret();
      $fieldResolvers = $interpreter->getFieldResolvers();
      $typeResolvers = $interpreter->getTypeResolvers();
      $hasLegacy = count(array_filter($autoload, fn ($entry) => array_key_exists('legacy', $entry) && $entry['legacy'])) > 0;
      if (!$hasLegacy) {
        $this->astCache?->set($resolversCache, [
          'field_resolvers' => $fieldResolvers,
          'type_resolvers' => $typeResolvers,
        ], CacheBackendInterface::CACHE_PERMANENT, ['graphql_schema']);
      }
    }

    foreach ($fieldResolvers as $type => $fields) {
      foreach ($fields as $field => $resolver) {
        $registry->addFieldResolver($type, $field, $resolver);
      }
    }

    foreach ($typeResolvers as $type => $resolver) {
      $registry->addTypeResolver($type, function ($value, $context, $info) use ($resolver) {
        $id = $resolver->resolve($value, [], $context, $info, new FieldContext($context, $info));
        // Ensure the type exists in the map before returning.
        return $this->typeMap[$id] ?? NULL;
      });
    }

    return $registry;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['schema_definition'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Schema definition'),
      '#default_value' => $this->configuration['schema_definition'],
      '#description' => $this->t(
        'Path to the schema definition file. Relative to webroot.'
      ),
    ];

    $form['autoload_contexts'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Autoload contexts'),
      '#default_value' => $this->configuration['autoload_contexts'],
      '#description' => $this->t(
        'Comma-separated list of context strings that are used to pick the correct directive implementations.'
      ),
    ];

    // Preserve config values coming from plugins (the ones that are not present on the form).
    foreach (array_keys($this->configuration) as $key) {
      if (!isset($form[$key])) {
        $form[$key] = [
          '#type' => 'value',
          '#value' => $this->configuration[$key],
        ];
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getSchemaDocument(array $extensions = []): DocumentNode {
    $document = parent::getSchemaDocument($extensions);
    foreach ($extensions as $extension) {
      if ($extension instanceof DirectableSchemaExtensionPluginBase) {
        $extension->setParentAst($document);
      }
    }
    return $document;
  }

  /**
   * {@inheritdoc}
   */
  public function getExtensionDocument(array $extensions = []): ?DocumentNode {
    return parent::getExtensionDocument($extensions);
  }

  /**
   * {@inheritdoc}
   */
  public function getExtensions(): array {
    return parent::getExtensions();
  }

}
