<?php

namespace Drupal\graphql_directives;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Symfony\Component\Yaml\Yaml;
use Webmozart\Glob\Glob;

/**
 * Service for aggregating GraphQL schema definition files.
 *
 * Searches for all directive definitions in enabled modules
 * and aggregates schema definitions from .graphqlrc.json files.
 */
class SchemaAggregator {

  public function __construct(
    protected ModuleHandlerInterface $moduleHandler,
  ) {}

  /**
   * Aggregate a new schema string based on a schema path.
   *
   * @param string $schemaPath
   *   The path to the schema definition.
   *
   * @return array{string}
   *   The list of files that are attached to this schema.
   */
  public function aggregate(string $schemaPath) : array {
    $root = DRUPAL_ROOT . '/' . $schemaPath;
    $files = [$root];
    foreach ($this->moduleHandler->getModuleDirectories() as $dir) {
      $directivesFile = $dir . '/directives.gql';
      if (file_exists($directivesFile)) {
        $files[] = $directivesFile;
      }
    }

    $configFile = $this->findConfigFile($root);
    if ($configFile) {
      $config = Yaml::parseFile($configFile);
      $schema = $config['schema'] ?? [];
      foreach ($schema as $pattern) {
        foreach (Glob::glob(dirname($configFile) . '/' . $pattern) as $file) {
          $files[] = $file;
        }
      }
    }

    return array_filter(array_values(array_unique(array_map('realpath', $files))));
  }

  /**
   * Scan for configuration files in any parent directories of the schema path.
   */
  protected function findConfigFile(string $schemaPath) : ?string {
    $current = dirname(realpath($schemaPath));
    $next = dirname($current);
    while ($next !== $current) {
      $file = realpath($current . '/.graphqlrc.json');
      if (file_exists($file)) {
        return $file;
      }
      $current = $next;
      $next = dirname($current);
    }
    return NULL;
  }

}
