<?php

namespace Drupal\Tests\graphql_directives\Kernel;

use Drupal\menu_link_content\Entity\MenuLinkContent;
use Drupal\menu_link_content\MenuLinkContentInterface;
use Drupal\Tests\graphql\Kernel\GraphQLTestBase;
use Drupal\Tests\graphql_directives\Traits\GraphQLDirectivesTestTrait;

/**
 * Test menu link GraphQL functionality.
 */
class MenuLinksTest extends GraphQLTestBase {
  use GraphQLDirectivesTestTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['graphql_directives'];

  /**
   * Count generated items for weight and labeling.
   *
   * @var int
   */
  protected $itemCount;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->itemCount = 0;
    $moduleDir = \Drupal::moduleHandler()->getModule('graphql_directives')->getPath();
    $this->setupDirectableSchema($moduleDir . '/tests/assets/menu_links');
  }

  /**
   * Helper function to create a menu item.
   */
  protected function createMenuItem(string $label, string $url, ?MenuLinkContentInterface $parent = NULL, string $menu = 'main') : MenuLinkContentInterface {
    $item = MenuLinkContent::create([
      'provider' => 'graphql_directives',
      'menu_name' => $menu,
      'title' => $label,
      'link' => ['uri' => $url],
      'weight' => $this->itemCount++,
    ] + ($parent ? ['parent' => $parent->getPluginId()] : []));
    $item->save();
    return $item;
  }

  /**
   * Empty menu test.
   */
  public function testEmptyMenu(): void {
    $metadata = $this->defaultCacheMetaData();
    $metadata->addCacheContexts(['languages:language_interface']);
    $metadata->addCacheTags(['config:system.menu.main']);
    $this->assertResults('{ menu { items { label } } }', [], [
      'menu' => [
        'items' => [],
      ],
    ], $metadata);
  }

  /**
   * Filled menu test.
   */
  public function testMenuItems() : void {
    $this->createMenuItem('Foo', 'internal:/foo');
    $this->createMenuItem('Bar', 'internal:/bar');
    $metadata = $this->defaultCacheMetaData();
    $metadata->addCacheContexts(['languages:language_interface']);
    $metadata->addCacheTags(['config:system.menu.main']);
    $this->assertResults('{ menu { items { label } } }', [], [
      'menu' => [
        'items' => [
          [
            'label' => 'Foo',
          ],
          [
            'label' => 'Bar',
          ],
        ],
      ],
    ], $metadata);
  }

  /**
   * Nested menu items test.
   */
  public function testMenuTree() : void {
    $foo = $this->createMenuItem('Foo', 'internal:/foo');
    $this->createMenuItem('Bar', 'internal:/bar', $foo);
    $metadata = $this->defaultCacheMetaData();
    $metadata->addCacheContexts(['languages:language_interface']);
    $metadata->addCacheTags(['config:system.menu.main']);
    $this->assertResults('{ menu { items { parent label } } }', [], [
      'menu' => [
        'items' => [
        ['parent' => $foo->getPluginId(), 'label' => 'Bar'],
        ['parent' => '', 'label' => 'Foo'],
        ],
      ],
    ], $metadata);
  }

  /**
   * Menu level restriction test.
   */
  public function testMenuLevels() : void {
    $foo = $this->createMenuItem('Foo', 'internal:/foo');
    $bar = $this->createMenuItem('Bar', 'internal:/bar', $foo);
    $this->createMenuItem('Baz', 'internal:/baz', $bar);
    $metadata = $this->defaultCacheMetaData();
    $metadata->addCacheContexts(['languages:language_interface']);
    $metadata->addCacheTags(['config:system.menu.main']);
    $this->assertResults('{ menu { items { label } } }', [], [
      'menu' => [
        'items' => [
        [
          'label' => 'Baz',
        ],
        [
          'label' => 'Bar',
        ],
        [
          'label' => 'Foo',
        ],
        ],
      ],
    ], $metadata);

    $this->assertResults('{ limited { items { label } } }', [], [
      'limited' => [
        'items' => [
        [
          'label' => 'Bar',
        ],
        [
          'label' => 'Foo',
        ],
        ],
      ],
    ], $metadata);
  }

  /**
   * Inaccessible items are filtered out.
   */
  public function testInaccessibleItems() : void {
    $this->createMenuItem('Foo', 'internal:/foo');
    // The admin link should not show up, since anonymous can't access it.
    $this->createMenuItem('Admin', 'internal:/admin');

    $metadata = $this->defaultCacheMetaData();
    $metadata->addCacheContexts(['languages:language_interface']);
    $metadata->addCacheTags(['config:system.menu.main']);
    $this->assertResults('{ menu { items { label } } }', [], [
      'menu' => [
        'items' => [
        [
          'label' => 'Foo',
        ],
        ],
      ],
    ], $metadata);
  }

  /**
   * Multilingual menus test.
   */
  public function testMultilingualMenu() : void {
    $this->container
      ->get('content_translation.manager')
      ->setEnabled('menu_link_content', 'menu_link_content', TRUE);

    $untranslated = $this->createMenuItem('English', 'internal:/untranslated');
    $translated = $this->createMenuItem('English', 'internal:/translated');
    $translation = $translated->addTranslation('de', [
      'title' => 'German',
    ]);
    $translation->save();

    $metadata = $this->defaultCacheMetaData();
    $metadata->addCacheContexts(['languages:language_interface']);
    $metadata->addCacheTags(['config:system.menu.main']);
    $this->assertResults('{ menu { items { label } } }', [], [
      'menu' => [
        'items' => [
          [
            'label' => $untranslated->label(),
          ],
          [
            'label' => $translated->label(),
          ],
        ],
      ],
    ], $metadata);

    $this->assertResults('{ menu:de { items { label } } }', [], [
      'menu' => [
        'items' => [
          [
            'label' => $translation->label(),
          ],
        ],
      ],
    ], $metadata);

    $this->assertResults('{ menu:fr { items { label } } }', [], [
      'menu' => [
        'items' => [],
      ],
    ], $metadata);
  }

}

