<?php

namespace Drupal\Tests\graphql_directives\Unit;

use Drupal\Component\DependencyInjection\ContainerInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\graphql_directives\ImplementationExtractor;
use Drupal\graphql_directives\MethodInterfaceException;
use Drupal\graphql_directives_test\TestResolvers;
use Prophecy\Prophecy\ObjectProphecy;

/**
 * @coversDefaultClass \Drupal\graphql_directives\ImplementationExtractor
 * @group graphql_directives
 */
class ImplementationExtractorTest extends UnitTestCase {

  protected ImplementationExtractor $extractor;

  /**
   *
   */
  protected function setUp() : void {
    parent::setUp();
    $containerProphecy = $this->prophesize(ContainerInterface::class);
    $containerProphecy->get('graphql_directives_test.test')->willReturn(new TestResolvers('static value'));
    $this->extractor = new ImplementationExtractor($containerProphecy->reveal());
  }

  /**
   * No directives.
   *
   * It should not break if the schema contains no directives.
   */
  public function testNoDirectives(): void {
    $this->assertEquals([], $this->extractor->extract(<<<EOT
      type Query {
        static: String
        service: String
        notImplemented: String
      }
    EOT));
  }

  /**
   * No implemented directives.
   *
   * It should return an empty list if there are no directives with
   * implementation hints.
   */
  public function testNoImplementedDirectives(): void {
    $this->assertEquals([], $this->extractor->extract(<<<EOT
      directive @whatever on FIELD_DEFINITION

      type Query {
        static: String @whatever
      }
    EOT));
  }

  /**
   * Service method.
   *
   * It should correctly resolve a service method implementation.
   */
  public function testServiceMethod(): void {
    $this->assertEquals([
      'example' => [
        'service' => 'graphql_directives_test.test',
        'method' => 'containerValue',
      ],
    ], $this->extractor->extract(<<<EOT
      """
      implementation(a, b): graphql_directives_test.test::containerValue
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a']));
  }

  /**
   * Static class method.
   *
   * It should correctly resolve a static class method implementation.
   */
  public function testStaticClassMethod(): void {
    $this->assertEquals([
      'example' => [
        'class' => '\\Drupal\\graphql_directives_test\\TestResolvers',
        'method' => 'staticValue',
      ],
    ], $this->extractor->extract(<<<EOT
      """
      implementation(a, b): \Drupal\graphql_directives_test\TestResolvers::staticValue
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a']));
  }

  /**
   * First implementation wins.
   *
   * It should select the first implementation in case of equal overlap.
   */
  public function testFirstImplementationWins(): void {
    $this->assertEquals([
      'example' => [
        'service' => 'graphql_directives_test.test',
        'method' => 'staticValue',
      ],
    ], $this->extractor->extract(<<<EOT
      """
      implementation(a, b): graphql_directives_test.test::staticValue
      implementation(a, b): graphql_directives_test.test::legacy
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a', 'b']));
  }

  /**
   * Biggest overlap wins.
   *
   * It should select the implementation with the biggest overlap of contexts.
   */
  public function testBiggestOverlapWins(): void {
    $this->assertEquals([
      'example' => [
        'service' => 'graphql_directives_test.test',
        'method' => 'legacy',
        'legacy' => TRUE,
      ],
    ], $this->extractor->extract(<<<EOT
      """
      implementation(a): graphql_directives_test.test::staticValue
      implementation(a, b): graphql_directives_test.test::legacy
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a', 'b']));
  }

  /**
   * No context matches.
   *
   * It should return an empty list if no contexts match.
   */
  public function testNoContextMatches(): void {
    $this->assertEquals([], $this->extractor->extract(<<<EOT
      """
      implementation(a, b): graphql_directives_test.test::legacy
      """
      directive @example on FIELD_DEFINITION
    EOT, ['c']));
  }

  /**
   * Verify that legacy implementations are marked as such.
   */
  public function testLegacy() : void {
    $this->assertEquals([
      'example' => [
        'service' => 'graphql_directives_test.test',
        'method' => 'legacy',
        'legacy' => TRUE,
      ],
    ], $this->extractor->extract(<<<EOT
      """
      @implementation(a): graphql_directives_test.test::legacy
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a']));
  }

  /**
   * Verify that an invalid argument type throws an exception.
   */
  public function testWrongArgumentType() : void {
    $this->expectExceptionObject(new MethodInterfaceException(TestResolvers::class, 'wrongArgumentType'));
    $this->extractor->extract(<<<EOT
      """
      @implementation(a): graphql_directives_test.test::wrongArgumentType
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a']);
  }

  /**
   * Verify that missing return types break.
   */
  public function testMissingReturnType() : void {
    $this->expectExceptionObject(new MethodInterfaceException(TestResolvers::class, 'missingReturnType'));
    $this->extractor->extract(<<<EOT
      """
      @implementation(a): graphql_directives_test.test::missingReturnType
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a']);
  }

  /**
   * Verify that wrong legacy return types break.
   */
  public function testLegacyWrongReturnType() : void {
    $this->expectExceptionObject(new MethodInterfaceException(TestResolvers::class, 'legacyWrongReturnType'));
    $this->extractor->extract(<<<EOT
      """
      @implementation(a): graphql_directives_test.test::legacyWrongReturnType
      """
      directive @example on FIELD_DEFINITION
    EOT, ['a']);
  }

}
